@@ -53,6 +53,7 @@ import signal
 import sys
 import token
 import tokenize
+import warnings
 
 import pep8
 
@@ -63,7 +64,7 @@ except NameError:
     unicode = str
 
 
-__version__ = '0.9.5'
+__version__ = '0.9.6'
 
 
 CR = '\r'
@@ -86,6 +87,28 @@ SHORTEN_OPERATOR_GROUPS = frozenset([
 DEFAULT_IGNORE = 'E24'
 
 
+# W602 is handled separately due to the need to avoid "with_traceback".
+CODE_TO_2TO3 = {
+    'W601': ['has_key'],
+    'W603': ['ne'],
+    'W604': ['repr'],
+    'W690': ['apply',
+             'except',
+             'exitfunc',
+             'idioms',
+             'import',
+             'numliterals',
+             'operator',
+             'paren',
+             'reduce',
+             'renames',
+             'standarderror',
+             'sys_exc',
+             'throw',
+             'tuple_params',
+             'xreadlines']}
+
+
 def open_with_encoding(filename, encoding=None, mode='r'):
     """Return opened file with a specific encoding."""
     if not encoding:
@@ -777,11 +800,13 @@ class FixPEP8(object):
             previous_line=(
                 self.source[line_index - 1] if line_index >= 1 else ''))
 
+        # Also sort alphabetically as a tie breaker (for determinism).
         candidates = list(sorted(
-            set(candidates).union([target]),
+            sorted(set(candidates).union([target])),
             key=lambda x: line_shortening_rank(x,
                                                self.newline,
-                                               self.indent_word)))
+                                               self.indent_word,
+                                               self.options.max_line_length)))
 
         if self.options.verbose >= 4:
             print(('-' * 79 + '\n').join([''] + candidates + ['']),
@@ -925,7 +950,7 @@ class FixPEP8(object):
         return range(1, 1 + original_length)
 
 
-def fix_e26(source, _=False):
+def fix_e26(source, aggressive=False, select='', ignore=''):
     """Format block comments."""
     if '#' not in source:
         # Optimization.
@@ -981,7 +1006,25 @@ def refactor(source, fixer_names, ignore=None):
     return new_text
 
 
-def fix_w602(source, aggressive=True):
+def code_to_2to3(select, ignore):
+    fixes = set()
+    for code, fix in CODE_TO_2TO3.items():
+        if code_match(code, select=select, ignore=ignore):
+            fixes |= set(fix)
+    return fixes
+
+
+def fix_w6(source, aggressive=True, select='', ignore=''):
+    """Fix various deprecated code (via lib2to3)."""
+    if not aggressive:
+        return source
+
+    return refactor(source,
+                    code_to_2to3(select=select,
+                                 ignore=ignore))
+
+
+def fix_w602(source, aggressive=True, select='', ignore=''):
     """Fix deprecated form of raising exception."""
     if not aggressive:
         return source
@@ -990,32 +1033,6 @@ def fix_w602(source, aggressive=True):
                     ignore='with_traceback')
 
 
-def fix_w6(source, aggressive=True):
-    """Fix various deprecated code (via lib2to3)."""
-    if not aggressive:
-        return source
-
-    return refactor(source,
-                    ['apply',
-                     'except',
-                     'exitfunc',
-                     'has_key',
-                     'idioms',
-                     'import',
-                     'ne',
-                     'numliterals',
-                     'operator',
-                     'paren',
-                     'reduce',
-                     'renames',
-                     'repr',
-                     'standarderror',
-                     'sys_exc',
-                     'throw',
-                     'tuple_params',
-                     'xreadlines'])
-
-
 def find_newline(source):
     """Return type of newline used in source."""
     cr, lf, crlf = 0, 0, 0
@@ -1150,7 +1167,6 @@ def _shorten_line(tokens, source, indentation, indent_word, newline,
         token_string = tkn[1]
 
         if (
-            aggressive and
             token_type == tokenize.COMMENT and
             not previous_line.rstrip().endswith('\\') and
             not source[tkn[2][1] + 1:].strip().lower().startswith(
@@ -1179,7 +1195,10 @@ def _shorten_line(tokens, source, indentation, indent_word, newline,
                 second_indent += indent_word
 
             second = (second_indent + source[offset:].lstrip())
-            if not second.strip():
+            if (
+                not second.strip() or
+                second.lstrip().startswith('#')
+            ):
                 continue
 
             # Do not begin a line with a comma
@@ -1761,7 +1780,7 @@ def code_match(code, select, ignore):
     return True
 
 
-def fix_string(source, options=None):
+def fix_code(source, options=None):
     """Return fixed source code."""
     if not options:
         options = parse_args([''])[0]
@@ -1773,6 +1792,13 @@ def fix_string(source, options=None):
     return fix_lines(sio.readlines(), options=options)
 
 
+def fix_string(source, options=None):
+    """Deprecated."""
+    warnings.warn('fix_string() is deprecated; use fix_code() instead',
+                  DeprecationWarning)
+    return fix_code(source, options)
+
+
 def fix_lines(source_lines, options, filename=''):
     """Return fixed source code."""
     tmp_source = ''.join(normalize_line_endings(source_lines))
@@ -1872,7 +1898,10 @@ def apply_global_fixes(source, options):
             if options.verbose:
                 print('--->  Applying global fix for {0}'.format(code.upper()),
                       file=sys.stderr)
-            source = function(source, options.aggressive)
+            source = function(source,
+                              aggressive=options.aggressive,
+                              select=options.select,
+                              ignore=options.ignore)
 
     return source
 
@@ -1924,8 +1953,8 @@ def create_parser():
                       help='enable non-whitespace changes; '
                            'multiple -a result in more aggressive changes')
     parser.add_option('--exclude', metavar='globs',
-                      help='exclude files that match these comma-separated '
-                           'globs')
+                      help='exclude file/directory names that match these '
+                           'comma-separated globs')
     parser.add_option('--list-fixes', action='store_true',
                       help='list codes for fixes; '
                            'used by --ignore and --select')
@@ -2050,7 +2079,7 @@ def docstring_summary(docstring):
     return docstring.split('\n')[0]
 
 
-def line_shortening_rank(candidate, newline, indent_word):
+def line_shortening_rank(candidate, newline, indent_word, max_line_length):
     """Return rank of candidate.
 
     This is for sorting candidates.
@@ -2061,14 +2090,22 @@ def line_shortening_rank(candidate, newline, indent_word):
         lines = candidate.split(newline)
 
         offset = 0
-        if lines[0].rstrip()[-1] not in '([{':
+        if (
+            not lines[0].lstrip().startswith('#') and
+            lines[0].rstrip()[-1] not in '([{'
+        ):
             for symbol in '([{':
                 offset = max(offset, 1 + lines[0].find(symbol))
 
-        max_length = max(offset + len(x.strip()) for x in lines)
-        rank += max_length
+        current_longest = max(offset + len(x.strip()) for x in lines)
+
+        rank += max(0, current_longest - max_line_length)
+
         rank += len(lines)
 
+        # Too much variation in line length is ugly.
+        rank += 2 * standard_deviation(len(line) for line in lines)
+
         bad_staring_symbol = {
             '(': ')',
             '[': ']',
@@ -2081,21 +2118,24 @@ def line_shortening_rank(candidate, newline, indent_word):
             ):
                 rank += 20
 
-        if re.match(r'.*[+\-\*/] \($', lines[0]):
-            # "1 * (\n" is ugly as hell.
-            rank += 100
-
         for current_line in lines:
+            if current_line.lstrip().startswith('#'):
+                continue
+
             for bad_start in ['.', '%', '+', '-', '/']:
                 if current_line.startswith(bad_start):
                     rank += 100
 
-            for ending in '([{':
+            if current_line.endswith(('(', '[', '{')):
                 # Avoid lonely opening. They result in longer lines.
-                if (
-                    current_line.endswith(ending) and
-                    len(current_line.strip()) <= len(indent_word)
-                ):
+                if len(current_line.strip()) <= len(indent_word):
+                    rank += 100
+
+                # Avoid ugliness of ", (\n".
+                if current_line[:-1].rstrip().endswith(','):
+                    rank += 100
+
+                if has_arithmetic_operator(current_line):
                     rank += 100
 
             if current_line.endswith('%'):
@@ -2112,6 +2152,25 @@ def line_shortening_rank(candidate, newline, indent_word):
     return max(0, rank)
 
 
+def standard_deviation(numbers):
+    """Return standard devation."""
+    numbers = list(numbers)
+    if not numbers:
+        return 0
+    mean = sum(numbers) / len(numbers)
+    return (sum((n - mean) ** 2 for n in numbers) /
+            len(numbers)) ** .5
+
+
+def has_arithmetic_operator(line):
+    """Return True if line contains any arithmetic operators."""
+    for operator in pep8.ARITHMETIC_OP:
+        if operator in line:
+            return True
+
+    return False
+
+
 def count_unbalanced_brackets(line):
     """Return number of unmatched open/close brackets."""
     count = 0
@@ -2162,14 +2221,16 @@ class LineEndingWrapper(object):
 
 def match_file(filename, exclude):
     """Return True if file is okay for modifying/recursing."""
-    if os.path.basename(filename).startswith('.'):
+    base_name = os.path.basename(filename)
+
+    if base_name.startswith('.'):
         return False
 
     for pattern in exclude:
-        if fnmatch.fnmatch(filename, pattern):
+        if fnmatch.fnmatch(base_name, pattern):
             return False
 
-    if not is_python_file(filename):
+    if not os.path.isdir(filename) and not is_python_file(filename):
         return False
 
     return True
@@ -2182,9 +2243,11 @@ def find_files(filenames, recursive, exclude):
         if recursive and os.path.isdir(name):
             for root, directories, children in os.walk(name):
                 filenames += [os.path.join(root, f) for f in children
-                              if match_file(f, exclude)]
+                              if match_file(os.path.join(root, f),
+                                            exclude)]
                 directories[:] = [d for d in directories
-                                  if not d.startswith('.')]
+                                  if match_file(os.path.join(root, d),
+                                                exclude)]
         else:
             yield name
 
@@ -2256,7 +2319,7 @@ def main():
 
             # LineEndingWrapper is unnecessary here due to the symmetry between
             # standard in and standard out.
-            sys.stdout.write(fix_string(sys.stdin.read(), options))
+            sys.stdout.write(fix_code(sys.stdin.read(), options))
         else:
             if options.in_place or options.diff:
                 filenames = list(set(args))
